#include "SampleGenerator.h"
#include "common.h"


using namespace std;
using namespace CHCDemoEngine;

HaltonSequence SphericalSampleGenerator::sHalton;
HaltonSequence PoissonDiscSampleGenerator::sHalton;
HaltonSequence PseudoRandomGenerator::sHalton;


SampleGenerator::SampleGenerator(int numSamples, float radius):
mNumSamples(numSamples), mRadius(radius)
{}


PoissonDiscSampleGenerator::PoissonDiscSampleGenerator(int numSamples, float radius):
SampleGenerator(numSamples, radius)
{}


void PoissonDiscSampleGenerator::Generate(float *samples) const
{
	// this is a hacky poisson sampling generator which does random dart-throwing
	// until it is not able to place any dart for a number of tries
	// in this case, the required min distance is reduced
	// the solution is a possion sampling with respect to the adjusted min distance
	// better solutions have been proposed, i.e., using hierarchical sampling

	const float maxTries = 1000;
	const float f_reduction = 0.9f;

	//static HaltonSequence halton;
	float r[2];

	// generates poisson distribution on disc
	float minDist = 2.0f * mRadius / sqrt((float)mNumSamples);

	//cout << "minDist before= " << minDist << endl;
	Sample2 *s = (Sample2 *)samples;

	for (int i = 0; i < mNumSamples; ++ i)
	{
		int tries = 0, totalTries = 0;

		// repeat until valid sample was found
		while (1)
		{
			++ tries;
			++ totalTries;

			r[0] = RandomValue(0, mRadius);
			r[1] = RandomValue(0, mRadius);
			//halton.GetNext(2, r);

			const float rx = r[0] * 2.0f - 1.0f;
			const float ry = r[1] * 2.0f - 1.0f;

			// check if in disk, else exit early
			if (rx * rx + ry * ry > mRadius * mRadius)
				continue;

			bool sampleValid = true;

			// check poisson property
			for (int j = 0; ((j < i) && sampleValid); ++ j)
			{
				const float dist = 
					sqrt((s[j].x - rx) * (s[j].x - rx) +
					     (s[j].y - ry) * (s[j].y - ry));
			
				if (dist < minDist)
					sampleValid = false;
			}

			if (sampleValid)
			{
				s[i].x = rx;
				s[i].y = ry;
				break;
			}

			if (tries > maxTries)
			{
				minDist *= f_reduction;
				tries = 0;
			}
		}
	}

	//cout << "minDist after= " << minDist << endl;
}


PseudoRandomGenerator::PseudoRandomGenerator(int numSamples, float radius):
SampleGenerator(numSamples, radius)
{}


void PseudoRandomGenerator::Generate(float *samples) const
{
	Sample2 *s = (Sample2 *)samples;

	int numSamples = 0;

	while (numSamples < mNumSamples)
	{
		const float rx = RandomValue(-mRadius, +mRadius);
		const float ry = RandomValue(-mRadius, +mRadius);

		// check if in disk, else exit early
		if (rx * rx + ry * ry > mRadius * mRadius)
			continue;

		s[numSamples].x = rx;
		s[numSamples].y = ry;

		++ numSamples;
	}
}


SphericalSampleGenerator::SphericalSampleGenerator(int numSamples, float radius):
SampleGenerator(numSamples, radius)
{}


void SphericalSampleGenerator::Generate(float *samples) const
{
	float r[2];
	Sample3 *s = (Sample3 *)samples;

	for (int i = 0; i < mNumSamples; ++ i)
	{
		r[0] = RandomValue(0, 1);
		r[1] = RandomValue(0, 1);
		//sHalton.GetNext(2, r);
		//r[0] = pSamples[i].x; r[1] = pSamples[i].y;

		// create stratified samples over sphere
		const float theta = 2.0f * acos(sqrt(1.0f - r[0]));
		const float phi = 2.0f * M_PI * r[1];
 
		s[i].x = mRadius * sin(theta) * cos(phi);
		s[i].y = mRadius * sin(theta) * sin(phi);
		s[i].z = mRadius * cos(theta);
	}

	//delete [] pSamples;
}